@@ -35,6 +35,7 @@
 
 #include "config.h"
 
+#include "attrs.h"
 #include "conf.h"
 #define P11_DEBUG_FLAG P11_DEBUG_RPC
 #include "debug.h"
@@ -58,10 +59,22 @@
 #include <string.h>
 #include <unistd.h>
 
+#ifdef ENABLE_NLS
+#include <libintl.h>
+#define _(x) dgettext(PACKAGE_NAME, x)
+#else
+#define _(x) (x)
+#endif
+
 /* The error returned on protocol failures */
 #define PARSE_ERROR CKR_DEVICE_ERROR
 #define PREP_ERROR  CKR_DEVICE_MEMORY
 
+typedef struct {
+	p11_virtual virt;
+	uint8_t version;
+} rpc_server;
+
 static CK_RV
 proto_read_byte_buffer (p11_rpc_message *msg,
                         CK_BYTE_PTR *buffer,
@@ -84,11 +97,17 @@ proto_read_byte_buffer (p11_rpc_message *msg,
 	*n_buffer = length;
 	*buffer = NULL;
 
+	/* length = -1 indicates length = 0, but buffer not NULL */
+	if (length == (uint32_t)-1) {
+		*n_buffer = 0;
+		length = 1; /*allocate 1 dummy byte */
+	}
+
 	/* If set to zero, then they just want the length */
 	if (length == 0)
 		return CKR_OK;
 
-	*buffer = p11_rpc_message_alloc_extra (msg, length * sizeof (CK_BYTE));
+	*buffer = p11_rpc_message_alloc_extra_array (msg, length, sizeof (CK_BYTE));
 	if (*buffer == NULL)
 		return CKR_DEVICE_MEMORY;
 
@@ -102,6 +121,7 @@ proto_read_byte_array (p11_rpc_message *msg,
 {
 	const unsigned char *data;
 	unsigned char valid;
+	uint32_t len;
 	size_t n_data;
 
 	assert (msg != NULL);
@@ -115,8 +135,10 @@ proto_read_byte_array (p11_rpc_message *msg,
 		return PARSE_ERROR;
 
 	if (!valid) {
+		if (!p11_rpc_buffer_get_uint32 (msg->input, &msg->parsed, &len))
+			return PARSE_ERROR;
 		*array = NULL;
-		*n_array = 0;
+		*n_array = len;
 		return CKR_OK;
 	}
 
@@ -186,7 +208,7 @@ proto_read_ulong_buffer (p11_rpc_message *msg,
 	if (length == 0)
 		return CKR_OK;
 
-	*buffer = p11_rpc_message_alloc_extra (msg, length * sizeof (CK_ULONG));
+	*buffer = p11_rpc_message_alloc_extra_array (msg, length, sizeof (CK_ULONG));
 	if (!*buffer)
 		return CKR_DEVICE_MEMORY;
 
@@ -225,28 +247,22 @@ proto_write_ulong_array (p11_rpc_message *msg,
 }
 
 static CK_RV
-proto_read_attribute_buffer (p11_rpc_message *msg,
-                             CK_ATTRIBUTE_PTR *result,
-                             CK_ULONG *n_result)
+proto_read_attribute_buffer_array (p11_rpc_message *msg,
+				   CK_ATTRIBUTE_PTR *result,
+				   CK_ULONG *n_result)
 {
-	CK_ATTRIBUTE_PTR attrs;
+	CK_RV rv;
+	CK_ATTRIBUTE_PTR attrs, array;
+	CK_ULONG n_array;
 	uint32_t n_attrs, i;
 	uint32_t value;
 
-	assert (msg != NULL);
-	assert (result != NULL);
-	assert (n_result != NULL);
-	assert (msg->input != NULL);
-
-	/* Make sure this is in the rigth order */
-	assert (!msg->signature || p11_rpc_message_verify_part (msg, "fA"));
-
 	/* Read the number of attributes */
 	if (!p11_rpc_buffer_get_uint32 (msg->input, &msg->parsed, &n_attrs))
 		return PARSE_ERROR;
 
 	/* Allocate memory for the attribute structures */
-	attrs = p11_rpc_message_alloc_extra (msg, n_attrs * sizeof (CK_ATTRIBUTE));
+	attrs = p11_rpc_message_alloc_extra_array (msg, n_attrs, sizeof (CK_ATTRIBUTE));
 	if (attrs == NULL)
 		return CKR_DEVICE_MEMORY;
 
@@ -266,6 +282,13 @@ proto_read_attribute_buffer (p11_rpc_message *msg,
 		if (value == 0) {
 			attrs[i].pValue = NULL;
 			attrs[i].ulValueLen = 0;
+		} else if (IS_ATTRIBUTE_ARRAY (attrs + i)) {
+			rv = proto_read_attribute_buffer_array (msg, &array, &n_array);
+			if (rv != CKR_OK)
+				return rv;
+			assert (n_array * sizeof (CK_ATTRIBUTE) <= value);
+			attrs[i].pValue = array;
+			attrs[i].ulValueLen = n_array * sizeof (CK_ATTRIBUTE);
 		} else {
 			attrs[i].pValue = p11_rpc_message_alloc_extra (msg, value);
 			if (!attrs[i].pValue)
@@ -279,6 +302,22 @@ proto_read_attribute_buffer (p11_rpc_message *msg,
 	return CKR_OK;
 }
 
+static CK_RV
+proto_read_attribute_buffer (p11_rpc_message *msg,
+			     CK_ATTRIBUTE_PTR *result,
+			     CK_ULONG *n_result)
+{
+	assert (msg != NULL);
+	assert (result != NULL);
+	assert (n_result != NULL);
+	assert (msg->input != NULL);
+
+	/* Make sure this is in the right order */
+	assert (!msg->signature || p11_rpc_message_verify_part (msg, "fA"));
+
+	return proto_read_attribute_buffer_array (msg, result, n_result);
+}
+
 static CK_RV
 proto_read_attribute_array (p11_rpc_message *msg,
                             CK_ATTRIBUTE_PTR *result,
@@ -292,7 +331,7 @@ proto_read_attribute_array (p11_rpc_message *msg,
 	assert (n_result != NULL);
 	assert (msg->input != NULL);
 
-	/* Make sure this is in the rigth order */
+	/* Make sure this is in the right order */
 	assert (!msg->signature || p11_rpc_message_verify_part (msg, "aA"));
 
 	/* Read the number of attributes */
@@ -300,39 +339,14 @@ proto_read_attribute_array (p11_rpc_message *msg,
 		return PARSE_ERROR;
 
 	/* Allocate memory for the attribute structures */
-	attrs = p11_rpc_message_alloc_extra (msg, n_attrs * sizeof (CK_ATTRIBUTE));
+	attrs = p11_rpc_message_alloc_extra_array (msg, n_attrs, sizeof (CK_ATTRIBUTE));
 	if (attrs == NULL)
 		return CKR_DEVICE_MEMORY;
 
 	/* Now go through and fill in each one */
-	for (i = 0; i < n_attrs; ++i) {
-		size_t offset = msg->parsed;
-		CK_ATTRIBUTE temp;
-
-		/* Check the length needed to store the value */
-		memset (&temp, 0, sizeof (temp));
-		if (!p11_rpc_buffer_get_attribute (msg->input, &offset, &temp)) {
-			msg->parsed = offset;
+	for (i = 0; i < n_attrs; ++i)
+		if (!p11_rpc_message_get_attribute (msg, msg->input, &msg->parsed, &attrs[i]))
 			return PARSE_ERROR;
-		}
-
-		attrs[i].type = temp.type;
-
-		/* Whether this one is valid or not */
-		if (temp.ulValueLen != ((CK_ULONG)-1)) {
-			size_t offset2 = msg->parsed;
-			attrs[i].pValue = p11_rpc_message_alloc_extra (msg, temp.ulValueLen);
-			if (!p11_rpc_buffer_get_attribute (msg->input, &offset2, &attrs[i])) {
-				msg->parsed = offset2;
-				return PARSE_ERROR;
-			}
-		} else {
-			attrs[i].pValue = NULL;
-			attrs[i].ulValueLen = -1;
-		}
-
-		msg->parsed = offset;
-	}
 
 	*result = attrs;
 	*n_result = n_attrs;
@@ -400,9 +414,40 @@ proto_read_null_string (p11_rpc_message *msg,
 	return CKR_OK;
 }
 
+static CK_RV
+proto_read_space_string (p11_rpc_message *msg,
+                         CK_UTF8CHAR_PTR *val,
+                         CK_ULONG len)
+{
+	const unsigned char *data;
+	size_t n_data;
+
+	assert (msg != NULL);
+	assert (val != NULL);
+	assert (msg->input != NULL);
+
+	/* Check that we're supposed to have this at this point */
+	assert (!msg->signature || p11_rpc_message_verify_part (msg, "s"));
+
+	if (!p11_rpc_buffer_get_byte_array (msg->input, &msg->parsed, &data, &n_data))
+		return PARSE_ERROR;
+
+	if (n_data != len)
+		return PARSE_ERROR;
+
+	/* Allocate a block of memory for it */
+	*val = p11_rpc_message_alloc_extra (msg, n_data);
+	if (*val == NULL)
+		return CKR_DEVICE_MEMORY;
+
+	memcpy (*val, data, n_data);
+
+	return CKR_OK;
+}
+
 static CK_RV
 proto_read_mechanism (p11_rpc_message *msg,
-                      CK_MECHANISM_PTR mech)
+                      CK_MECHANISM_PTR *mech)
 {
 	size_t offset;
 	CK_MECHANISM temp;
@@ -422,19 +467,30 @@ proto_read_mechanism (p11_rpc_message *msg,
 		return PARSE_ERROR;
 	}
 
-	mech->mechanism = temp.mechanism;
+	/*
+	 * The NULL mechanism is used for C_*Init () functions to
+	 * cancel operation.  We use a special value 0xffffffff as a
+	 * marker to indicate that.
+	 */
+	if (temp.mechanism == 0xffffffff) {
+		*mech = NULL;
+		msg->parsed = offset;
+		return CKR_OK;
+	}
+
+	(*mech)->mechanism = temp.mechanism;
 
 	/* The mechanism doesn't require parameter */
 	if (temp.ulParameterLen == 0) {
-		mech->pParameter = NULL;
-		mech->ulParameterLen = 0;
+		(*mech)->pParameter = NULL;
+		(*mech)->ulParameterLen = 0;
 		msg->parsed = offset;
 		return CKR_OK;
 	}
 
 	/* Actually retrieve the parameter */
-	mech->pParameter = p11_rpc_message_alloc_extra (msg, temp.ulParameterLen);
-	if (!p11_rpc_buffer_get_mechanism (msg->input, &msg->parsed, mech))
+	(*mech)->pParameter = p11_rpc_message_alloc_extra (msg, temp.ulParameterLen);
+	if (!p11_rpc_buffer_get_mechanism (msg->input, &msg->parsed, *mech))
 		return PARSE_ERROR;
 
 	assert (msg->parsed == offset);
@@ -548,7 +604,7 @@ call_ready (p11_rpc_message *msg)
 	 */
 
 	if (p11_buffer_failed (msg->output)) {
-		p11_message ("invalid request from module, probably too short"); \
+		p11_message (_("invalid request from module, probably too short")); \
 		return PARSE_ERROR;
 	}
 
@@ -558,7 +614,7 @@ call_ready (p11_rpc_message *msg)
 	msg->input = NULL;
 
 	if (!p11_rpc_message_prep (msg, msg->call_id, P11_RPC_RESPONSE)) {
-		p11_message ("couldn't initialize rpc response");
+		p11_message (_("couldn't initialize rpc response"));
 		return CKR_DEVICE_MEMORY;
 	}
 
@@ -597,10 +653,14 @@ call_ready (p11_rpc_message *msg)
 	if (!p11_rpc_message_read_ulong (msg, &val)) \
 		{ _ret = PARSE_ERROR; goto _cleanup; }
 
-#define IN_STRING(val) \
+#define IN_ZERO_STRING(val) \
 	_ret = proto_read_null_string (msg, &val); \
 	if (_ret != CKR_OK) goto _cleanup;
 
+#define IN_SPACE_STRING(val, len) \
+	_ret = proto_read_space_string (msg, &val, len); \
+	if (_ret != CKR_OK) goto _cleanup;
+
 #define IN_BYTE_BUFFER(buffer, buffer_len) \
 	_ret = proto_read_byte_buffer (msg, &buffer, &buffer_len); \
 	if (_ret != CKR_OK) goto _cleanup;
@@ -688,9 +748,10 @@ rpc_C_Initialize (CK_X_FUNCTION_LIST *self,
 	if (ret == CKR_OK) {
 
 		/* Check to make sure the header matches */
-		if (n_handshake != P11_RPC_HANDSHAKE_LEN ||
+		if (!handshake ||
+		    n_handshake != P11_RPC_HANDSHAKE_LEN ||
 		    memcmp (handshake, P11_RPC_HANDSHAKE, n_handshake) != 0) {
-			p11_message ("invalid handshake received from connecting module");
+			p11_message (_("invalid handshake received from connecting module"));
 			ret = CKR_GENERAL_ERROR;
 		}
 	}
@@ -833,7 +894,24 @@ rpc_C_InitToken (CK_X_FUNCTION_LIST *self,
 	BEGIN_CALL (InitToken);
 		IN_ULONG (slot_id);
 		IN_BYTE_ARRAY (pin, pin_len);
-		IN_STRING (label);
+		IN_ZERO_STRING (label);
+	PROCESS_CALL ((self, slot_id, pin, pin_len, label));
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_InitToken2 (CK_X_FUNCTION_LIST *self,
+                  p11_rpc_message *msg)
+{
+	CK_SLOT_ID slot_id;
+	CK_UTF8CHAR_PTR pin;
+	CK_ULONG pin_len;
+	CK_UTF8CHAR_PTR label;
+
+	BEGIN_CALL (InitToken);
+		IN_ULONG (slot_id);
+		IN_BYTE_ARRAY (pin, pin_len);
+		IN_SPACE_STRING (label, 32);
 	PROCESS_CALL ((self, slot_id, pin, pin_len, label));
 	END_CALL;
 }
@@ -1156,14 +1234,15 @@ rpc_C_EncryptInit (CK_X_FUNCTION_LIST *self,
                    p11_rpc_message *msg)
 {
 	CK_SESSION_HANDLE session;
-	CK_MECHANISM mechanism;
+	CK_MECHANISM mechanism_;
+	CK_MECHANISM_PTR mechanism = &mechanism_;
 	CK_OBJECT_HANDLE key;
 
 	BEGIN_CALL (EncryptInit);
 		IN_ULONG (session);
 		IN_MECHANISM (mechanism);
 		IN_ULONG (key);
-	PROCESS_CALL ((self, session, &mechanism, key));
+	PROCESS_CALL ((self, session, mechanism, key));
 	END_CALL;
 
 }
@@ -1227,14 +1306,15 @@ rpc_C_DecryptInit (CK_X_FUNCTION_LIST *self,
                     p11_rpc_message *msg)
 {
 	CK_SESSION_HANDLE session;
-	CK_MECHANISM mechanism;
+	CK_MECHANISM mechanism_;
+	CK_MECHANISM_PTR mechanism = &mechanism_;
 	CK_OBJECT_HANDLE key;
 
 	BEGIN_CALL (DecryptInit);
 		IN_ULONG (session);
 		IN_MECHANISM (mechanism);
 		IN_ULONG (key);
-	PROCESS_CALL ((self, session, &mechanism, key));
+	PROCESS_CALL ((self, session, mechanism, key));
 	END_CALL;
 }
 
@@ -1297,12 +1377,13 @@ rpc_C_DigestInit (CK_X_FUNCTION_LIST *self,
                   p11_rpc_message *msg)
 {
 	CK_SESSION_HANDLE session;
-	CK_MECHANISM mechanism;
+	CK_MECHANISM mechanism_;
+	CK_MECHANISM_PTR mechanism = &mechanism_;
 
 	BEGIN_CALL (DigestInit);
 		IN_ULONG (session);
 		IN_MECHANISM (mechanism);
-	PROCESS_CALL ((self, session, &mechanism));
+	PROCESS_CALL ((self, session, mechanism));
 	END_CALL;
 }
 
@@ -1375,14 +1456,15 @@ rpc_C_SignInit (CK_X_FUNCTION_LIST *self,
                 p11_rpc_message *msg)
 {
 	CK_SESSION_HANDLE session;
-	CK_MECHANISM mechanism;
+	CK_MECHANISM mechanism_;
+	CK_MECHANISM_PTR mechanism = &mechanism_;
 	CK_OBJECT_HANDLE key;
 
 	BEGIN_CALL (SignInit);
 		IN_ULONG (session);
 		IN_MECHANISM (mechanism);
 		IN_ULONG (key);
-	PROCESS_CALL ((self, session, &mechanism, key));
+	PROCESS_CALL ((self, session, mechanism, key));
 	END_CALL;
 }
 
@@ -1442,14 +1524,15 @@ rpc_C_SignRecoverInit (CK_X_FUNCTION_LIST *self,
                        p11_rpc_message *msg)
 {
 	CK_SESSION_HANDLE session;
-	CK_MECHANISM mechanism;
+	CK_MECHANISM mechanism_;
+	CK_MECHANISM_PTR mechanism = &mechanism_;
 	CK_OBJECT_HANDLE key;
 
 	BEGIN_CALL (SignRecoverInit);
 		IN_ULONG (session);
 		IN_MECHANISM (mechanism);
 		IN_ULONG (key);
-	PROCESS_CALL ((self, session, &mechanism, key));
+	PROCESS_CALL ((self, session, mechanism, key));
 	END_CALL;
 }
 
@@ -1477,14 +1560,15 @@ rpc_C_VerifyInit (CK_X_FUNCTION_LIST *self,
                   p11_rpc_message *msg)
 {
 	CK_SESSION_HANDLE session;
-	CK_MECHANISM mechanism;
+	CK_MECHANISM mechanism_;
+	CK_MECHANISM_PTR mechanism = &mechanism_;
 	CK_OBJECT_HANDLE key;
 
 	BEGIN_CALL (VerifyInit);
 		IN_ULONG (session);
 		IN_MECHANISM (mechanism);
 		IN_ULONG (key);
-	PROCESS_CALL ((self, session, &mechanism, key));
+	PROCESS_CALL ((self, session, mechanism, key));
 	END_CALL;
 }
 
@@ -1541,14 +1625,15 @@ rpc_C_VerifyRecoverInit (CK_X_FUNCTION_LIST *self,
                          p11_rpc_message *msg)
 {
 	CK_SESSION_HANDLE session;
-	CK_MECHANISM mechanism;
+	CK_MECHANISM mechanism_;
+	CK_MECHANISM_PTR mechanism = &mechanism_;
 	CK_OBJECT_HANDLE key;
 
 	BEGIN_CALL (VerifyRecoverInit);
 		IN_ULONG (session);
 		IN_MECHANISM (mechanism);
 		IN_ULONG (key);
-	PROCESS_CALL ((self, session, &mechanism, key));
+	PROCESS_CALL ((self, session, mechanism, key));
 	END_CALL;
 }
 
@@ -1652,7 +1737,8 @@ rpc_C_GenerateKey (CK_X_FUNCTION_LIST *self,
                    p11_rpc_message *msg)
 {
 	CK_SESSION_HANDLE session;
-	CK_MECHANISM mechanism;
+	CK_MECHANISM mechanism_;
+	CK_MECHANISM_PTR mechanism = &mechanism_;
 	CK_ATTRIBUTE_PTR template;
 	CK_ULONG count;
 	CK_OBJECT_HANDLE key;
@@ -1661,7 +1747,7 @@ rpc_C_GenerateKey (CK_X_FUNCTION_LIST *self,
 		IN_ULONG (session);
 		IN_MECHANISM (mechanism);
 		IN_ATTRIBUTE_ARRAY (template, count);
-	PROCESS_CALL ((self, session, &mechanism, template, count, &key));
+	PROCESS_CALL ((self, session, mechanism, template, count, &key));
 		OUT_ULONG (key);
 	END_CALL;
 }
@@ -1671,7 +1757,8 @@ rpc_C_GenerateKeyPair (CK_X_FUNCTION_LIST *self,
                        p11_rpc_message *msg)
 {
 	CK_SESSION_HANDLE session;
-	CK_MECHANISM mechanism;
+	CK_MECHANISM mechanism_;
+	CK_MECHANISM_PTR mechanism = &mechanism_;
 	CK_ATTRIBUTE_PTR public_key_template;
 	CK_ULONG public_key_attribute_count;
 	CK_ATTRIBUTE_PTR private_key_template;
@@ -1684,7 +1771,8 @@ rpc_C_GenerateKeyPair (CK_X_FUNCTION_LIST *self,
 		IN_MECHANISM (mechanism);
 		IN_ATTRIBUTE_ARRAY (public_key_template, public_key_attribute_count);
 		IN_ATTRIBUTE_ARRAY (private_key_template, private_key_attribute_count);
-	PROCESS_CALL ((self, session, &mechanism, public_key_template, public_key_attribute_count, private_key_template, private_key_attribute_count, &public_key, &private_key));
+	PROCESS_CALL ((self, session, mechanism, public_key_template, public_key_attribute_count,
+	               private_key_template, private_key_attribute_count, &public_key, &private_key));
 		OUT_ULONG (public_key);
 		OUT_ULONG (private_key);
 	END_CALL;
@@ -1695,7 +1783,8 @@ rpc_C_WrapKey (CK_X_FUNCTION_LIST *self,
                p11_rpc_message *msg)
 {
 	CK_SESSION_HANDLE session;
-	CK_MECHANISM mechanism;
+	CK_MECHANISM mechanism_;
+	CK_MECHANISM_PTR mechanism = &mechanism_;
 	CK_OBJECT_HANDLE wrapping_key;
 	CK_OBJECT_HANDLE key;
 	CK_BYTE_PTR wrapped_key;
@@ -1707,7 +1796,7 @@ rpc_C_WrapKey (CK_X_FUNCTION_LIST *self,
 		IN_ULONG (wrapping_key);
 		IN_ULONG (key);
 		IN_BYTE_BUFFER (wrapped_key, wrapped_key_len);
-	PROCESS_CALL ((self, session, &mechanism, wrapping_key, key, wrapped_key, &wrapped_key_len));
+	PROCESS_CALL ((self, session, mechanism, wrapping_key, key, wrapped_key, &wrapped_key_len));
 		OUT_BYTE_ARRAY (wrapped_key, wrapped_key_len);
 	END_CALL;
 }
@@ -1717,7 +1806,8 @@ rpc_C_UnwrapKey (CK_X_FUNCTION_LIST *self,
                  p11_rpc_message *msg)
 {
 	CK_SESSION_HANDLE session;
-	CK_MECHANISM mechanism;
+	CK_MECHANISM mechanism_;
+	CK_MECHANISM_PTR mechanism = &mechanism_;
 	CK_OBJECT_HANDLE unwrapping_key;
 	CK_BYTE_PTR wrapped_key;
 	CK_ULONG wrapped_key_len;
@@ -1731,7 +1821,7 @@ rpc_C_UnwrapKey (CK_X_FUNCTION_LIST *self,
 		IN_ULONG (unwrapping_key);
 		IN_BYTE_ARRAY (wrapped_key, wrapped_key_len);
 		IN_ATTRIBUTE_ARRAY (template, attribute_count);
-	PROCESS_CALL ((self, session, &mechanism, unwrapping_key, wrapped_key, wrapped_key_len, template, attribute_count, &key));
+	PROCESS_CALL ((self, session, mechanism, unwrapping_key, wrapped_key, wrapped_key_len, template, attribute_count, &key));
 		OUT_ULONG (key);
 	END_CALL;
 }
@@ -1741,7 +1831,8 @@ rpc_C_DeriveKey (CK_X_FUNCTION_LIST *self,
                  p11_rpc_message *msg)
 {
 	CK_SESSION_HANDLE session;
-	CK_MECHANISM mechanism;
+	CK_MECHANISM mechanism_;
+	CK_MECHANISM_PTR mechanism = &mechanism_;
 	CK_OBJECT_HANDLE base_key;
 	CK_ATTRIBUTE_PTR template;
 	CK_ULONG attribute_count;
@@ -1752,7 +1843,7 @@ rpc_C_DeriveKey (CK_X_FUNCTION_LIST *self,
 		IN_MECHANISM (mechanism);
 		IN_ULONG (base_key);
 		IN_ATTRIBUTE_ARRAY (template, attribute_count);
-	PROCESS_CALL ((self, session, &mechanism, base_key, template, attribute_count, &key));
+	PROCESS_CALL ((self, session, mechanism, base_key, template, attribute_count, &key));
 		OUT_ULONG (key);
 	END_CALL;
 }
@@ -1788,6 +1879,416 @@ rpc_C_GenerateRandom (CK_X_FUNCTION_LIST *self,
 	END_CALL;
 }
 
+static CK_RV
+rpc_C_LoginUser (CK_X_FUNCTION_LIST *self,
+                 p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+	CK_USER_TYPE user_type;
+	CK_UTF8CHAR_PTR pin;
+	CK_ULONG pin_len;
+	CK_UTF8CHAR_PTR username;
+	CK_ULONG username_len;
+
+	BEGIN_CALL (LoginUser);
+		IN_ULONG (session);
+		IN_ULONG (user_type);
+		IN_BYTE_ARRAY (pin, pin_len);
+		IN_BYTE_ARRAY (username, username_len);
+	PROCESS_CALL ((self, session, user_type, pin, pin_len, username, username_len));
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_SessionCancel (CK_X_FUNCTION_LIST *self,
+                     p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+	CK_FLAGS flags;
+
+	BEGIN_CALL (SessionCancel);
+		IN_ULONG (session);
+		IN_ULONG (flags);
+	PROCESS_CALL ((self, session, flags));
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_MessageEncryptInit (CK_X_FUNCTION_LIST *self,
+                          p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+	CK_MECHANISM mechanism_;
+	CK_MECHANISM_PTR mechanism = &mechanism_;
+	CK_OBJECT_HANDLE key;
+
+	BEGIN_CALL (MessageEncryptInit);
+		IN_ULONG (session);
+		IN_MECHANISM (mechanism);
+		IN_ULONG (key);
+	PROCESS_CALL ((self, session, mechanism, key));
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_EncryptMessage (CK_X_FUNCTION_LIST *self,
+                      p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+	CK_BYTE_PTR parameter;
+	CK_ULONG parameter_len;
+	CK_BYTE_PTR associated_data;
+	CK_ULONG associated_data_len;
+	CK_BYTE_PTR plaintext;
+	CK_ULONG plaintext_len;
+	CK_BYTE_PTR ciphertext;
+	CK_ULONG ciphertext_len;
+
+	BEGIN_CALL (EncryptMessage);
+		IN_ULONG (session);
+		IN_BYTE_ARRAY (parameter, parameter_len);
+		IN_BYTE_ARRAY (associated_data, associated_data_len);
+		IN_BYTE_ARRAY (plaintext, plaintext_len);
+		IN_BYTE_BUFFER (ciphertext, ciphertext_len);
+	PROCESS_CALL ((self, session, (void *)parameter, parameter_len, associated_data, associated_data_len,
+	               plaintext, plaintext_len, ciphertext, &ciphertext_len));
+		OUT_BYTE_ARRAY (ciphertext, ciphertext_len);
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_EncryptMessageBegin (CK_X_FUNCTION_LIST *self,
+                           p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+	CK_BYTE_PTR parameter;
+	CK_ULONG parameter_len;
+	CK_BYTE_PTR associated_data;
+	CK_ULONG associated_data_len;
+
+	BEGIN_CALL (EncryptMessageBegin);
+		IN_ULONG (session)
+		IN_BYTE_ARRAY (parameter, parameter_len)
+		IN_BYTE_ARRAY (associated_data, associated_data_len)
+	PROCESS_CALL ((self, session, (void *)parameter, parameter_len, associated_data, associated_data_len));
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_EncryptMessageNext (CK_X_FUNCTION_LIST *self,
+                          p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+	CK_BYTE_PTR parameter;
+	CK_ULONG parameter_len;
+	CK_BYTE_PTR plaintext_part;
+	CK_ULONG plaintext_part_len;
+	CK_BYTE_PTR ciphertext_part;
+	CK_ULONG ciphertext_part_len;
+	CK_FLAGS flags;
+
+	BEGIN_CALL (EncryptMessageNext);
+		IN_ULONG (session);
+		IN_BYTE_ARRAY (parameter, parameter_len);
+		IN_BYTE_ARRAY (plaintext_part, plaintext_part_len);
+		IN_BYTE_BUFFER (ciphertext_part, ciphertext_part_len);
+		IN_ULONG (flags);
+	PROCESS_CALL ((self, session, (void *)parameter, parameter_len, plaintext_part, plaintext_part_len,
+	               ciphertext_part, &ciphertext_part_len, flags));
+		OUT_BYTE_ARRAY (ciphertext_part, ciphertext_part_len)
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_MessageEncryptFinal (CK_X_FUNCTION_LIST *self,
+                           p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+
+	BEGIN_CALL (MessageEncryptFinal);
+		IN_ULONG (session);
+	PROCESS_CALL ((self, session));
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_MessageDecryptInit (CK_X_FUNCTION_LIST *self,
+                          p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+	CK_MECHANISM mechanism_;
+	CK_MECHANISM_PTR mechanism = &mechanism_;
+	CK_OBJECT_HANDLE key;
+
+	BEGIN_CALL (MessageDecryptInit);
+		IN_ULONG (session);
+		IN_MECHANISM (mechanism);
+		IN_ULONG (key);
+	PROCESS_CALL ((self, session, mechanism, key));
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_DecryptMessage (CK_X_FUNCTION_LIST *self,
+                      p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+	CK_BYTE_PTR parameter;
+	CK_ULONG parameter_len;
+	CK_BYTE_PTR associated_data;
+	CK_ULONG associated_data_len;
+	CK_BYTE_PTR ciphertext;
+	CK_ULONG ciphertext_len;
+	CK_BYTE_PTR plaintext;
+	CK_ULONG plaintext_len;
+
+	BEGIN_CALL (DecryptMessage);
+		IN_ULONG (session);
+		IN_BYTE_ARRAY (parameter, parameter_len);
+		IN_BYTE_ARRAY (associated_data, associated_data_len);
+		IN_BYTE_ARRAY (ciphertext, ciphertext_len);
+		IN_BYTE_BUFFER (plaintext, plaintext_len);
+	PROCESS_CALL ((self, session, (void *)parameter, parameter_len, associated_data, associated_data_len,
+	               ciphertext, ciphertext_len, plaintext, &plaintext_len));
+		OUT_BYTE_ARRAY (plaintext, plaintext_len);
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_DecryptMessageBegin (CK_X_FUNCTION_LIST *self,
+                           p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+	CK_BYTE_PTR parameter;
+	CK_ULONG parameter_len;
+	CK_BYTE_PTR associated_data;
+	CK_ULONG associated_data_len;
+
+	BEGIN_CALL (DecryptMessageBegin);
+		IN_ULONG (session);
+		IN_BYTE_ARRAY (parameter, parameter_len);
+		IN_BYTE_ARRAY (associated_data, associated_data_len);
+	PROCESS_CALL ((self, session, (void *)parameter, parameter_len, associated_data, associated_data_len));
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_DecryptMessageNext (CK_X_FUNCTION_LIST *self,
+                          p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+	CK_BYTE_PTR parameter;
+	CK_ULONG parameter_len;
+	CK_BYTE_PTR ciphertext_part;
+	CK_ULONG ciphertext_part_len;
+	CK_BYTE_PTR plaintext_part;
+	CK_ULONG plaintext_part_len;
+	CK_FLAGS flags;
+
+	BEGIN_CALL (DecryptMessageNext);
+		IN_ULONG (session);
+		IN_BYTE_ARRAY (parameter, parameter_len);
+		IN_BYTE_ARRAY (ciphertext_part, ciphertext_part_len);
+		IN_BYTE_BUFFER (plaintext_part, plaintext_part_len);
+		IN_ULONG (flags);
+	PROCESS_CALL ((self, session, (void *)parameter, parameter_len, ciphertext_part, ciphertext_part_len,
+	               plaintext_part, &plaintext_part_len, flags));
+		OUT_BYTE_ARRAY (plaintext_part, plaintext_part_len);
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_MessageDecryptFinal (CK_X_FUNCTION_LIST *self,
+                           p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+
+	BEGIN_CALL (MessageDecryptFinal);
+		IN_ULONG (session);
+	PROCESS_CALL ((self, session));
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_MessageSignInit (CK_X_FUNCTION_LIST *self,
+                       p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+	CK_MECHANISM mechanism_;
+	CK_MECHANISM_PTR mechanism = &mechanism_;
+	CK_OBJECT_HANDLE key;
+
+	BEGIN_CALL (MessageSignInit);
+		IN_ULONG (session);
+		IN_MECHANISM (mechanism);
+		IN_ULONG (key);
+	PROCESS_CALL ((self, session, mechanism, key));
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_SignMessage (CK_X_FUNCTION_LIST *self,
+                   p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+	CK_BYTE_PTR parameter;
+	CK_ULONG parameter_len;
+	CK_BYTE_PTR data;
+	CK_ULONG data_len;
+	CK_BYTE_PTR signature;
+	CK_ULONG signature_len;
+
+	BEGIN_CALL (SignMessage);
+		IN_ULONG (session);
+		IN_BYTE_ARRAY (parameter, parameter_len);
+		IN_BYTE_ARRAY (data, data_len);
+		IN_BYTE_BUFFER (signature, signature_len);
+	PROCESS_CALL ((self, session, (void *)parameter, parameter_len, data, data_len,
+	               signature, &signature_len));
+		OUT_BYTE_ARRAY (signature, signature_len);
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_SignMessageBegin (CK_X_FUNCTION_LIST *self,
+                        p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+	CK_BYTE_PTR parameter;
+	CK_ULONG parameter_len;
+
+	BEGIN_CALL (SignMessageBegin);
+		IN_ULONG (session);
+		IN_BYTE_ARRAY (parameter, parameter_len);
+	PROCESS_CALL ((self, session, (void *)parameter, parameter_len));
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_SignMessageNext (CK_X_FUNCTION_LIST *self,
+                       p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+	CK_BYTE_PTR parameter;
+	CK_ULONG parameter_len;
+	CK_BYTE_PTR data;
+	CK_ULONG data_len;
+	CK_BBOOL signature_expected;
+	CK_BYTE_PTR signature;
+	CK_ULONG signature_len;
+
+	BEGIN_CALL (SignMessageNext);
+		IN_ULONG (session);
+		IN_BYTE_ARRAY (parameter, parameter_len);
+		IN_BYTE_ARRAY (data, data_len);
+		IN_BYTE (signature_expected);
+		IN_BYTE_BUFFER (signature, signature_len);
+	PROCESS_CALL ((self, session, (void *)parameter, parameter_len, data, data_len,
+		       (signature_expected ? signature : NULL),
+		       (signature_expected ? &signature_len : NULL)));
+		OUT_BYTE_ARRAY (signature, (signature_expected ? signature_len : 0));
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_MessageSignFinal (CK_X_FUNCTION_LIST *self,
+                        p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+
+	BEGIN_CALL (MessageSignFinal);
+		IN_ULONG (session);
+	PROCESS_CALL ((self, session));
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_MessageVerifyInit (CK_X_FUNCTION_LIST *self,
+                         p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+	CK_MECHANISM mechanism_;
+	CK_MECHANISM_PTR mechanism = &mechanism_;
+	CK_OBJECT_HANDLE key;
+
+	BEGIN_CALL (MessageVerifyInit);
+		IN_ULONG (session);
+		IN_MECHANISM (mechanism);
+		IN_ULONG (key);
+	PROCESS_CALL ((self, session, mechanism, key));
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_VerifyMessage (CK_X_FUNCTION_LIST *self,
+                     p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+	CK_BYTE_PTR parameter;
+	CK_ULONG parameter_len;
+	CK_BYTE_PTR data;
+	CK_ULONG data_len;
+	CK_BYTE_PTR signature;
+	CK_ULONG signature_len;
+
+	BEGIN_CALL (VerifyMessage);
+		IN_ULONG (session);
+		IN_BYTE_ARRAY (parameter, parameter_len);
+		IN_BYTE_ARRAY (data, data_len);
+		IN_BYTE_ARRAY (signature, signature_len);
+	PROCESS_CALL ((self, session, (void *)parameter, parameter_len, data, data_len,
+	               signature, signature_len));
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_VerifyMessageBegin (CK_X_FUNCTION_LIST *self,
+                          p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+	CK_BYTE_PTR parameter;
+	CK_ULONG parameter_len;
+
+	BEGIN_CALL (VerifyMessageBegin);
+		IN_ULONG (session);
+		IN_BYTE_ARRAY (parameter, parameter_len);
+	PROCESS_CALL ((self, session, (void *)parameter, parameter_len));
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_VerifyMessageNext (CK_X_FUNCTION_LIST *self,
+                         p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+	CK_BYTE_PTR parameter;
+	CK_ULONG parameter_len;
+	CK_BYTE_PTR data;
+	CK_ULONG data_len;
+	CK_BYTE_PTR signature;
+	CK_ULONG signature_len;
+
+	BEGIN_CALL (VerifyMessageNext);
+		IN_ULONG (session);
+		IN_BYTE_ARRAY (parameter, parameter_len);
+		IN_BYTE_ARRAY (data, data_len);
+		IN_BYTE_ARRAY (signature, signature_len);
+	PROCESS_CALL ((self, session, (void *)parameter, parameter_len, data, data_len, signature, signature_len));
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_MessageVerifyFinal (CK_X_FUNCTION_LIST *self,
+                          p11_rpc_message *msg)
+{
+	CK_SESSION_HANDLE session;
+
+	BEGIN_CALL (MessageVerifyFinal);
+		IN_ULONG (session);
+	PROCESS_CALL ((self, session));
+	END_CALL;
+}
+
 bool
 p11_rpc_server_handle (CK_X_FUNCTION_LIST *self,
                        p11_buffer *request,
@@ -1807,7 +2308,7 @@ p11_rpc_server_handle (CK_X_FUNCTION_LIST *self,
 
 	if (!p11_rpc_message_parse (&msg, P11_RPC_REQUEST)) {
 		p11_rpc_message_clear (&msg);
-		p11_message ("couldn't parse pkcs11 rpc message");
+		p11_message (_("couldn't parse pkcs11 rpc message"));
 		return false;
 	}
 
@@ -1886,6 +2387,31 @@ p11_rpc_server_handle (CK_X_FUNCTION_LIST *self,
 	CASE_CALL (C_SeedRandom)
 	CASE_CALL (C_GenerateRandom)
 	CASE_CALL (C_WaitForSlotEvent)
+	/* PKCS #11 3.0 */
+	CASE_CALL (C_LoginUser)
+	CASE_CALL (C_SessionCancel)
+	CASE_CALL (C_MessageEncryptInit)
+	CASE_CALL (C_EncryptMessage)
+	CASE_CALL (C_EncryptMessageBegin)
+	CASE_CALL (C_EncryptMessageNext)
+	CASE_CALL (C_MessageEncryptFinal)
+	CASE_CALL (C_MessageDecryptInit)
+	CASE_CALL (C_DecryptMessage)
+	CASE_CALL (C_DecryptMessageBegin)
+	CASE_CALL (C_DecryptMessageNext)
+	CASE_CALL (C_MessageDecryptFinal)
+	CASE_CALL (C_MessageSignInit)
+	CASE_CALL (C_SignMessage)
+	CASE_CALL (C_SignMessageBegin)
+	CASE_CALL (C_SignMessageNext)
+	CASE_CALL (C_MessageSignFinal)
+	CASE_CALL (C_MessageVerifyInit)
+	CASE_CALL (C_VerifyMessage)
+	CASE_CALL (C_VerifyMessageBegin)
+	CASE_CALL (C_VerifyMessageNext)
+	CASE_CALL (C_MessageVerifyFinal)
+
+	CASE_CALL (C_InitToken2)
 	#undef CASE_CALL
 	default:
 		/* This should have been caught by the parse code */
@@ -1894,7 +2420,7 @@ p11_rpc_server_handle (CK_X_FUNCTION_LIST *self,
 	};
 
 	if (p11_buffer_failed (msg.output)) {
-		p11_message ("out of memory error putting together message");
+		p11_message (_("out of memory error putting together message"));
 		p11_rpc_message_clear (&msg);
 		return false;
 	}
@@ -1918,7 +2444,7 @@ p11_rpc_server_handle (CK_X_FUNCTION_LIST *self,
 		if (!p11_rpc_message_prep (&msg, P11_RPC_CALL_ERROR, P11_RPC_RESPONSE) ||
 		    !p11_rpc_message_write_ulong (&msg, (uint32_t)ret) ||
 		    p11_buffer_failed (msg.output)) {
-			p11_message ("out of memory responding with error");
+			p11_message (_("out of memory responding with error"));
 			p11_rpc_message_clear (&msg);
 			return false;
 		}
@@ -1944,8 +2470,7 @@ p11_kit_remote_serve_module (CK_FUNCTION_LIST *module,
                              int out_fd)
 {
 	p11_rpc_status status;
-	unsigned char version;
-	p11_virtual virt;
+	rpc_server server;
 	p11_buffer options;
 	p11_buffer buffer;
 	size_t state;
@@ -1957,28 +2482,33 @@ p11_kit_remote_serve_module (CK_FUNCTION_LIST *module,
 	p11_buffer_init (&options, 0);
 	p11_buffer_init (&buffer, 0);
 
-	p11_virtual_init (&virt, &p11_virtual_base, module, NULL);
+	p11_virtual_init (&server.virt, &p11_virtual_base, module, NULL);
 
-	switch (read (in_fd, &version, 1)) {
+	switch (read (in_fd, &server.version, 1)) {
 	case 0:
 		goto out;
 	case 1:
-		if (version != 0) {
-			p11_message ("unsupported version received: %d", (int)version);
+#if P11_RPC_PROTOCOL_VERSION_MINIMUM > 0
+		if (server.version < P11_RPC_PROTOCOL_VERSION_MINIMUM) {
+			p11_message (_("unsupported version received: %d"), (int)version);
 			goto out;
 		}
+#endif
 		break;
 	default:
-		p11_message_err (errno, "couldn't read credential byte");
+		p11_message_err (errno, _("couldn't read credential byte"));
 		goto out;
 	}
 
-	version = 0;
-	switch (write (out_fd, &version, 1)) {
+	if (server.version > P11_RPC_PROTOCOL_VERSION_MAXIMUM) {
+		server.version = P11_RPC_PROTOCOL_VERSION_MAXIMUM;
+	}
+
+	switch (write (out_fd, &server.version, 1)) {
 	case 1:
 		break;
 	default:
-		p11_message_err (errno, "couldn't write credential byte");
+		p11_message_err (errno, _("couldn't write credential byte"));
 		goto out;
 	}
 
@@ -2001,12 +2531,12 @@ p11_kit_remote_serve_module (CK_FUNCTION_LIST *module,
 		case P11_RPC_AGAIN:
 			assert_not_reached ();
 		case P11_RPC_ERROR:
-			p11_message_err (errno, "failed to read rpc message");
+			p11_message_err (errno, _("failed to read rpc message"));
 			goto out;
 		}
 
-		if (!p11_rpc_server_handle (&virt.funcs, &buffer, &buffer)) {
-			p11_message ("unexpected error handling rpc message");
+		if (!p11_rpc_server_handle (&server.virt.funcs, &buffer, &buffer)) {
+			p11_message (_("unexpected error handling rpc message"));
 			goto out;
 		}
 
@@ -2024,7 +2554,7 @@ p11_kit_remote_serve_module (CK_FUNCTION_LIST *module,
 		case P11_RPC_AGAIN:
 			assert_not_reached ();
 		case P11_RPC_ERROR:
-			p11_message_err (errno, "failed to write rpc message");
+			p11_message_err (errno, _("failed to write rpc message"));
 			goto out;
 		}
 	}
@@ -2033,7 +2563,7 @@ out:
 	p11_buffer_uninit (&buffer);
 	p11_buffer_uninit (&options);
 
-	p11_virtual_uninit (&virt);
+	p11_virtual_uninit (&server.virt);
 
 	return ret;
 }
@@ -2219,14 +2749,22 @@ p11_kit_remote_serve_tokens (const char **tokens,
 		filter = p11_dict_get (filters, module);
 		if (filter == NULL) {
 			lower = calloc (1, sizeof (p11_virtual));
+			if (lower == NULL) {
+				error = ENOMEM;
+				goto out;
+			}
 			p11_virtual_init (lower, &p11_virtual_base, module, NULL);
-			filter = p11_filter_subclass (lower, NULL);
+			filter = p11_filter_subclass (lower, free);
 			if (filter == NULL) {
 				error = EINVAL;
 				p11_message_err (error, "couldn't subclass filter");
 				goto out;
 			}
-			p11_dict_set (filters, module, filter);
+			if (!p11_dict_set (filters, module, filter)) {
+				error = EINVAL;
+				p11_message_err (error, "couldn't register filter");
+				goto out;
+			}
 		}
 
 		for (i = 0; i < n_tokens; i++) {
@@ -2241,8 +2779,7 @@ p11_kit_remote_serve_tokens (const char **tokens,
 	filtered = p11_array_new ((p11_destroyer)module_unwrap);
 	p11_dict_iterate (filters, &filters_iter);
 	while (p11_dict_next (&filters_iter, NULL, &value)) {
-		module = p11_virtual_wrap ((p11_virtual *)value,
-					   (p11_destroyer)p11_virtual_uninit);
+		module = p11_virtual_wrap ((p11_virtual *)value, NULL);
 		if (module == NULL) {
 			error = EINVAL;
 			p11_message_err (error, "couldn't wrap filter module");
@@ -2281,6 +2818,11 @@ p11_kit_remote_serve_tokens (const char **tokens,
 		p11_kit_modules_release (modules);
 	if (error != 0)
 		errno = error;
+	if (uris) {
+		for (i = 0; i < n_tokens; i++)
+			p11_kit_uri_free (uris[i]);
+		free (uris);
+	}
 
 	return ret;
 }
