@@ -40,7 +40,7 @@
 #include "path.h"
 #include "p11-kit.h"
 #include "remote.h"
-#include "tool.h"
+#include "options.h"
 
 #include <assert.h>
 #include <errno.h>
@@ -60,12 +60,18 @@
 #include <sys/types.h>
 #include <sys/un.h>
 #include <sys/wait.h>
+#include <sys/stat.h>
 #include <unistd.h>
 
 #ifdef WITH_SYSTEMD
 #include <systemd/sd-daemon.h>
 #endif
 
+#ifdef HAVE_VSOCK
+#include "vsock.h"
+#include <linux/vm_sockets.h>
+#endif
+
 #ifdef HAVE_SIGHANDLER_T
 #define SIGHANDLER_T sighandler_t
 #elif HAVE_SIG_T
@@ -79,6 +85,13 @@ typedef void (*sighandler_t)(int);
 
 #endif /* OS_UNIX */
 
+#ifdef ENABLE_NLS
+#include <libintl.h>
+#define _(x) dgettext(PACKAGE_NAME, x)
+#else
+#define _(x) (x)
+#endif
+
 typedef struct {
 	const char **tokens;
 	size_t n_tokens;
@@ -90,6 +103,11 @@ typedef struct {
 	uid_t uid;
 	gid_t gid;
 
+#ifdef HAVE_VSOCK
+	unsigned int vsock_cid;
+	unsigned int vsock_port;
+#endif /* HAVE_VSOCK */
+
 	int socket;
 #endif /* OS_UNIX */
 
@@ -138,6 +156,30 @@ server_new (const char **tokens, size_t n_tokens, const char *provider,
 	server->socket_name = socket_name;
 
 #ifdef OS_UNIX
+	if (strncmp (socket_name, "unix:path=", 10) == 0) {
+		server->socket_name = socket_name + 10;
+#ifdef HAVE_VSOCK
+	} else if (strncmp (socket_name, "vsock:", 6) == 0) {
+		if (!p11_vsock_parse_addr(socket_name + 6,
+					  &server->vsock_cid,
+					  &server->vsock_port)) {
+			p11_message (_("failed to parse vsock address: '%s'"),
+				     socket_name + 6);
+			free (server);
+			return NULL;
+		}
+		if (server->vsock_cid == VMADDR_CID_ANY) {
+			/* We need to print the right CID so that clients can
+			 * know the address to connect to. Just binding to
+			 * VMADDR_CID_ANY isn't stunningly useful unless the
+			 * client knows the CID through other means */
+			p11_vsock_get_local_cid (&server->vsock_cid);
+			/* On error, we'll just report 0xFFFFFFFF */
+		}
+		server->socket_name = NULL;
+#endif /* HAVE_VSOCK */
+	}
+
 	server->socket = -1;
 #endif /* OS_UNIX */
 
@@ -170,7 +212,7 @@ static bool need_children_cleanup = false;
 static bool terminate = false;
 static unsigned children_avail = 0;
 static bool quiet = false;
-static bool csh = false;
+static bool csh_opt = false;
 
 #define P11_KIT_SERVER_ADDRESS_ENV "P11_KIT_SERVER_ADDRESS"
 #define P11_KIT_SERVER_PID_ENV "P11_KIT_SERVER_PID"
@@ -199,9 +241,9 @@ cleanup_children (void)
 			children_avail--;
 		if (WIFSIGNALED (status)) {
 			if (WTERMSIG (status) == SIGSEGV)
-				p11_message ("child %u died with sigsegv", (unsigned)pid);
+				p11_message (_("child %u died with sigsegv"), (unsigned)pid);
 			else
-				p11_message ("child %u died with signal %d", (unsigned)pid, (int)WTERMSIG (status));
+				p11_message (_("child %u died with signal %d"), (unsigned)pid, (int)WTERMSIG (status));
 		}
 	}
 	need_children_cleanup = false;
@@ -255,18 +297,19 @@ exec_external (int argc,
 }
 
 static int
-create_socket (const char *address,
-	       uid_t uid,
-	       gid_t gid)
+create_unix_socket (const char *address,
+		    uid_t uid,
+		    gid_t gid)
 {
 	int rc, sd;
 	struct sockaddr_un sa;
 	const char *socket_file;
+	mode_t socket_mask;
 
 	memset (&sa, 0, sizeof(sa));
 	sa.sun_family = AF_UNIX;
 
-	return_val_if_fail (strlen (address) < sizeof (sa.sun_path), -1);
+	return_val_if_fail (strlen (address) < sizeof (sa.sun_path) - 1, -1);
 	strncpy (sa.sun_path, address, sizeof (sa.sun_path));
 	socket_file = sa.sun_path;
 
@@ -274,30 +317,33 @@ create_socket (const char *address,
 
 	sd = socket (AF_UNIX, SOCK_STREAM, 0);
 	if (sd == -1) {
-		p11_message_err (errno, "could not create socket %s", socket_file);
+		p11_message_err (errno, _("could not create socket %s"), socket_file);
 		return -1;
 	}
 
-	umask (066);
+	socket_mask = S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
+	if (gid != -1)
+		socket_mask &= ~(S_IRGRP | S_IWGRP);
+	umask (socket_mask);
 	rc = bind (sd, (struct sockaddr *)&sa, SUN_LEN (&sa));
 	if (rc == -1) {
 		close (sd);
-		p11_message_err (errno, "could not bind socket %s", socket_file);
+		p11_message_err (errno, _("could not bind socket %s"), socket_file);
 		return -1;
 	}
 
 	rc = listen (sd, 1024);
 	if (rc == -1) {
 		close (sd);
-		p11_message_err (errno, "could not listen to socket %s", socket_file);
+		p11_message_err (errno, _("could not listen to socket %s"), socket_file);
 		return 1;
 	}
 
-	if (uid != -1 && gid != -1) {
+	if (uid != -1 || gid != -1) {
 		rc = chown (socket_file, uid, gid);
 		if (rc == -1) {
 			close (sd);
-			p11_message_err (errno, "could not chown socket %s", socket_file);
+			p11_message_err (errno, _("could not chown socket %s"), socket_file);
 			return -1;
 		}
 	}
@@ -305,6 +351,43 @@ create_socket (const char *address,
 	return sd;
 }
 
+#ifdef HAVE_VSOCK
+static int
+create_vsock_socket (unsigned int cid,
+		     unsigned int port)
+{
+	int rc, sd;
+	struct sockaddr_vm sa;
+
+	memset (&sa, 0, sizeof(sa));
+	sa.svm_family = AF_VSOCK;
+	sa.svm_cid = cid;
+	sa.svm_port = port;
+
+	sd = socket (AF_VSOCK, SOCK_STREAM, 0);
+	if (sd == -1) {
+		p11_message_err (errno, _("could not create socket %u:%u"), cid, port);
+		return -1;
+	}
+
+	rc = bind (sd, (struct sockaddr *)&sa, sizeof(sa));
+	if (rc == -1) {
+		close (sd);
+		p11_message_err (errno, _("could not bind socket %u:%u"), cid, port);
+		return -1;
+	}
+
+	rc = listen (sd, 1024);
+	if (rc == -1) {
+		close (sd);
+		p11_message_err (errno, _("could not listen to socket %u:%u"), cid, port);
+		return 1;
+	}
+
+	return sd;
+}
+#endif /* HAVE_VSOCK */
+
 static bool
 check_credentials (int fd,
 		   uid_t uid,
@@ -316,20 +399,20 @@ check_credentials (int fd,
 
 	rc = p11_get_upeer_id (fd, &tuid, &tgid, NULL);
 	if (rc == -1) {
-		p11_message_err (errno, "could not check uid from socket");
+		p11_message_err (errno, _("could not check uid from socket"));
 		close (fd);
 		return false;
 	}
 
 	if (uid != -1 && uid != tuid) {
-		p11_message ("connecting uid (%u) doesn't match expected (%u)",
+		p11_message (_("connecting uid (%u) doesn't match expected (%u)"),
 			     (unsigned)tuid, (unsigned)uid);
 		close (fd);
 		return false;
 	}
 
 	if (gid != -1 && gid != tgid) {
-		p11_message ("connecting gid (%u) doesn't match expected (%u)",
+		p11_message (_("connecting gid (%u) doesn't match expected (%u)"),
 			     (unsigned)tgid, (unsigned)gid);
 		close (fd);
 		return false;
@@ -339,14 +422,26 @@ check_credentials (int fd,
 }
 
 static bool
-print_environment (pid_t pid, const char *socket_path, bool csh)
+print_environment (pid_t pid, Server *server, bool csh)
 {
 	char *path, *address;
-	int rc;
+	int rc = -1;
 
-	path = p11_path_encode (socket_path);
-	rc = asprintf (&address, "unix:path=%s", path);
-	free (path);
+	if (server->socket_name) {
+		path = p11_path_encode (server->socket_name);
+		rc = asprintf (&address, "unix:path=%s", path);
+		free (path);
+#ifdef HAVE_VSOCK
+	} else if (server->vsock_cid || server->vsock_port) {
+		if (server->vsock_cid == VMADDR_CID_ANY) {
+			rc = asprintf (&address, "vsock:port=%u",
+				       server->vsock_port);
+		} else {
+			rc = asprintf (&address, "vsock:cid=%u;port=%u",
+				       server->vsock_cid, server->vsock_port);
+		}
+#endif
+	}
 	if (rc < 0)
 		return false;
 	if (csh) {
@@ -398,7 +493,7 @@ server_loop (Server *server,
 	if (!foreground) {
 		pid = fork ();
 		if (pid == -1) {
-			p11_message_err (errno, "could not fork() to daemonize");
+			p11_message_err (errno, _("could not fork() to daemonize"));
 			return 1;
 		}
 		if (pid == 0) {
@@ -406,12 +501,12 @@ server_loop (Server *server,
 			close (STDOUT_FILENO);
 		}
 		if (pid != 0) {
-			if (!print_environment (pid, server->socket_name, csh))
+			if (!print_environment (pid, server, csh_opt))
 				return 1;
 			exit (0);
 		}
 		if (setsid () == -1) {
-			p11_message_err (errno, "could not create a new session");
+			p11_message_err (errno, _("could not create a new session"));
 			return 1;
 		}
 	}
@@ -419,24 +514,28 @@ server_loop (Server *server,
 #ifdef WITH_SYSTEMD
 	ret = sd_listen_fds (0);
 	if (ret > 1) {
-		p11_message ("too many file descriptors received");
+		p11_message (_("too many file descriptors received"));
 		return 1;
 	} else if (ret == 1) {
 		server->socket = SD_LISTEN_FDS_START + 0;
 	} else
 #endif
-	{
-		server->socket = create_socket (server->socket_name, server->uid, server->gid);
-		if (server->socket == -1)
-			return 1;
+	if (server->socket_name) {
+		server->socket = create_unix_socket (server->socket_name, server->uid, server->gid);
+#ifdef HAVE_VSOCK
+	} else if (server->vsock_cid || server->vsock_port) {
+		server->socket = create_vsock_socket (server->vsock_cid, server->vsock_port);
+#endif
 	}
+	if (server->socket == -1)
+		return 1;
 
 	sigprocmask (SIG_BLOCK, &blockset, NULL);
 
 	/* for testing purposes, even when started in foreground,
 	 * print the envvars */
 	if (foreground) {
-		if (!print_environment (getpid (), server->socket_name, csh))
+		if (!print_environment (getpid (), server, csh_opt))
 			return 1;
 		fflush (stdout);
 	}
@@ -459,7 +558,7 @@ server_loop (Server *server,
 
 		/* timeout */
 		if (ret == 0 && children_avail == 0 && timeout != NULL) {
-			p11_message ("no connections to %s for %lu secs, exiting", server->socket_name, timeout->tv_sec);
+			p11_message (_("no connections to %s for %lu secs, exiting"), server->socket_name, timeout->tv_sec);
 			break;
 		}
 
@@ -468,17 +567,18 @@ server_loop (Server *server,
 			cfd = accept (server->socket, (struct sockaddr *)&sa, &sa_len);
 			if (cfd == -1) {
 				if (errno != EINTR)
-					p11_message_err (errno, "could not accept from socket %s", server->socket_name);
+					p11_message_err (errno, _("could not accept from socket %s"), server->socket_name);
 				continue;
 			}
 
-			if (!check_credentials (cfd, server->uid, server->gid))
+			if (server->socket_name &&
+			    !check_credentials (cfd, server->uid, server->gid))
 				continue;
 
 			pid = fork ();
 			switch (pid) {
 			case -1:
-				p11_message_err (errno, "failed to fork for accept");
+				p11_message_err (errno, _("failed to fork for accept"));
 				continue;
 			/* Child */
 			case 0:
@@ -546,7 +646,6 @@ main (int argc,
 	const struct passwd *pwd;
 	const struct group *grp;
 	bool foreground = false;
-	bool csh_opt = false;
 	bool kill_opt = false;
 	struct timespec *timeout = NULL, ts;
 	char *name = NULL;
@@ -624,7 +723,7 @@ main (int argc,
 		case opt_group:
 			grp = getgrnam (optarg);
 			if (grp == NULL) {
-				p11_message ("unknown group: %s", optarg);
+				p11_message (_("unknown group: %s"), optarg);
 				return 2;
 			}
 			gid = grp->gr_gid;
@@ -632,7 +731,7 @@ main (int argc,
 		case opt_user:
 			pwd = getpwnam (optarg);
 			if (pwd == NULL) {
-				p11_message ("unknown user: %s", optarg);
+				p11_message (_("unknown user: %s"), optarg);
 				return 2;
 			}
 			uid = pwd->pw_uid;
@@ -640,7 +739,7 @@ main (int argc,
 		case opt_run_as_group:
 			grp = getgrnam (optarg);
 			if (grp == NULL) {
-				p11_message ("unknown group: %s", optarg);
+				p11_message (_("unknown group: %s"), optarg);
 				return 2;
 			}
 			run_as_gid = grp->gr_gid;
@@ -648,7 +747,7 @@ main (int argc,
 		case opt_run_as_user:
 			pwd = getpwnam (optarg);
 			if (pwd == NULL) {
-				p11_message ("unknown user: %s", optarg);
+				p11_message (_("unknown user: %s"), optarg);
 				return 2;
 			}
 			run_as_uid = pwd->pw_uid;
@@ -663,12 +762,10 @@ main (int argc,
 			kill_opt = true;
 			break;
 		case opt_csh:
-			csh = true;
 			csh_opt = true;
 			break;
 		case opt_sh:
-			csh = false;
-			csh_opt = true;
+			csh_opt = false;
 			break;
 		case opt_help:
 		case '?':
@@ -688,12 +785,12 @@ main (int argc,
 		return 2;
 	}
 
-	if (!csh_opt) {
+	if (!opt_sh && !opt_csh) {
 		const char *shell = secure_getenv ("SHELL");
 		size_t len;
 		if (shell != NULL && (len = strlen (shell)) > 2 &&
 		    strncmp (shell + len - 3, "csh", 3) == 0)
-			csh = true;
+			csh_opt = true;
 	}
 
 	if (kill_opt) {
@@ -717,7 +814,7 @@ main (int argc,
 			exit (1);
 		}
 
-		if (csh) {
+		if (csh_opt) {
 			printf ("unsetenv %s;\n",
 				P11_KIT_SERVER_ADDRESS_ENV);
 			printf ("unsetenv %s;\n",
@@ -733,13 +830,13 @@ main (int argc,
 
 	if (run_as_gid != -1) {
 		if (setgid (run_as_gid) == -1) {
-			p11_message_err (errno, "cannot set gid to %u", (unsigned)run_as_gid);
+			p11_message_err (errno, _("cannot set gid to %u"), (unsigned)run_as_gid);
 			ret = 1;
 			goto out;
 		}
 
 		if (setgroups (1, &run_as_gid) == -1) {
-			p11_message_err (errno, "cannot setgroups to %u", (unsigned)run_as_gid);
+			p11_message_err (errno, _("cannot setgroups to %u"), (unsigned)run_as_gid);
 			ret = 1;
 			goto out;
 		}
@@ -747,7 +844,7 @@ main (int argc,
 
 	if (run_as_uid != -1) {
 		if (setuid (run_as_uid) == -1) {
-			p11_message_err (errno, "cannot set uid to %u", (unsigned)run_as_uid);
+			p11_message_err (errno, _("cannot set uid to %u"), (unsigned)run_as_uid);
 			ret = 1;
 			goto out;
 		}
@@ -763,7 +860,7 @@ main (int argc,
 
 		runtime_dir = secure_getenv ("XDG_RUNTIME_DIR");
 		if (!runtime_dir || !runtime_dir[0]) {
-			p11_message_err (errno, "cannot determine runtime directory");
+			p11_message_err (errno, _("cannot determine runtime directory"));
 			ret = 1;
 			goto out;
 		}
@@ -775,7 +872,7 @@ main (int argc,
 		}
 
 		if (mkdir (socket_base, 0700) == -1 && errno != EEXIST) {
-			p11_message_err (errno, "cannot create %s", socket_base);
+			p11_message_err (errno, _("cannot create %s"), socket_base);
 			ret = 1;
 			goto out;
 		}
@@ -1046,7 +1143,7 @@ main (int argc,
 	};
 
 	if (!load_windows_functions ()) {
-		p11_message ("couldn't initialize Windows security functions");
+		p11_message (_("couldn't initialize Windows security functions"));
 		return 1;
 	}
 
@@ -1207,7 +1304,7 @@ get_sids (void)
 	if (!user_sid) {
 		user_sid = get_user_sid ();
 		if (user_sid == NULL) {
-			p11_message ("unable to construct SID for %s: %lu",
+			p11_message (_("unable to construct SID for %s: %lu"),
 				     "current user",
 				     GetLastError ());
 			return false;
@@ -1219,7 +1316,7 @@ get_sids (void)
 					       SECURITY_WORLD_RID,
 					       0, 0, 0, 0, 0, 0, 0,
 					       &world_sid)) {
-			p11_message ("unable to construct SID for %s: %lu",
+			p11_message (_("unable to construct SID for %s: %lu"),
 				     "world",
 				     GetLastError ());
 			return false;
@@ -1231,7 +1328,7 @@ get_sids (void)
 					       SECURITY_NETWORK_RID,
 					       0, 0, 0, 0, 0, 0, 0,
 					       &network_sid)) {
-			p11_message ("unable to construct SID for %s: %lu",
+			p11_message (_("unable to construct SID for %s: %lu"),
 				     "local same-user access only",
 				     GetLastError ());
 			return false;
@@ -1274,31 +1371,31 @@ make_private_security_descriptor (DWORD permissions,
 
 	acl_err = SetEntriesInAclA (3, ea, NULL, acl);
 	if (acl_err != ERROR_SUCCESS || *acl == NULL) {
-		p11_message ("unable to construct ACL: %d", acl_err);
+		p11_message (_("unable to construct ACL: %d"), acl_err);
 		goto cleanup;
 	}
 
 	*psd = (PSECURITY_DESCRIPTOR) LocalAlloc (LPTR, SECURITY_DESCRIPTOR_MIN_LENGTH);
 	if (!*psd) {
-		p11_message ("unable to allocate security descriptor: %lu",
+		p11_message (_("unable to allocate security descriptor: %lu"),
 			     GetLastError ());
 		goto cleanup;
 	}
 
 	if (!InitializeSecurityDescriptor (*psd, SECURITY_DESCRIPTOR_REVISION)) {
-		p11_message ("unable to initialise security descriptor: %lu",
+		p11_message (_("unable to initialise security descriptor: %lu"),
 			     GetLastError ());
 		goto cleanup;
 	}
 
 	if (!SetSecurityDescriptorOwner (*psd, user_sid, FALSE)) {
-		p11_message ("unable to set owner in security descriptor: %lu",
+		p11_message (_("unable to set owner in security descriptor: %lu"),
 			     GetLastError ());
 		goto cleanup;
 	}
 
 	if (!SetSecurityDescriptorDacl (*psd, TRUE, *acl, FALSE)) {
-		p11_message ("unable to set DACL in security descriptor: %lu",
+		p11_message (_("unable to set DACL in security descriptor: %lu"),
 			     GetLastError ());
 		goto cleanup;
 	}
